import json
import asyncio
import platform
import datetime


class Host():
    """
        Class to represent a Host to be pinged.

        Example:
        ```
        Host('localhost', db)
        ```

        db is optional, but without it
        database functionallity will not
        function.
    """

    def __init__(self, addr, db = None):
        self.__db = db
        self.__addr = addr
        self.__alive = False
        self.__timestamp = datetime.datetime.utcnow()
        self.__errors = {}
        self.__data = None

        asyncio.ensure_future(self.__ping_loop())

    def __str__(self):
        return json.dumps(self.get_json(), indent=4, default=str)

    def __eq__(self, other):
        if self.get_addr() != other.get_addr():
            return False
        if self.is_alive() != other.is_alive():
            return False
        if self.get_timestamp() != other.get_timestamp():
            return False

        return True
    
    def get_csv(self):
        """
            Return Host as a string of csv

            Example:
            ```
            print(Host('localhost', db).get_csv())
            ```
        """
        
        return f"{self.__addr}, {self.__alive}, {self.__timestamp}\n"
    
    def get_json(self):
        """
            Return Host as a dict of json data

            Example:
            ```
            print(Host('localhost', db).get_json())
            ```
        """

        return {
            'addr': self.__addr,
            'alive': self.__alive,
            'timestamp': self.__timestamp,
            'errors': self.__errors,
            'data': self.__data,
        }

    def get_timestamp(self):
        return self.__timestamp

    def get_addr(self):
        return self.__addr

    def is_alive(self):
        return self.__alive

    def __check_error(self, errors, query, message):
        if self.__data.find(query) != -1:
            errors.append(message)

    async def __ping_loop(self):
        while True:
            asyncio.ensure_future(self.ping())
            await asyncio.sleep(1)

    async def ping(self):
        """
            Ping the provided self.
            
            Example:
            ```
            print(ping('localhost', db))
            ```
        """

        param = '-c'
        if platform.system().lower() == 'windows':
            param = '-n'

        proc = await asyncio.subprocess.create_subprocess_exec(
            'ping',
            '-w', '1',
            param, '1',
            self.__addr,
            stdout=asyncio.subprocess.PIPE,
        )
        proc_data = await proc.communicate()
        self.__data = proc_data[0].decode()

        errors = []
        self.__check_error(errors, '100%', 'host_packet_loss')
        self.__check_error(errors, 'Request timed out', 'host_timed_out')
        self.__check_error(errors, 'Name or service not known', 'host_name_not_found')
        self.__check_error(errors, 'Destination host unreachable', 'host_unreachable')
        self.__check_error(errors, 'Ping request could not find host', 'host_not_found')

        if len(errors) == 0:
            self.__alive = True
        else:
            self.__alive = False
            self.__db.insert_error(self.get_json())
        
        for error in errors:
            self.__errors[error] = True

        self.__timestamp = datetime.datetime.utcnow()



"""
    Entrypoint for running the CLI script

    Example:
    ```
    aiocheck.cli.main(['localhost'])
    ```
"""

import aiocheck

import os
import sys
import asyncio
import argparse


def main(addresses = []):
    """
        Entrypoint for running the CLI script

        Example:
        ```
        aiocheck.cli.main(['localhost'])
        ```
    """

    parser = argparse.ArgumentParser(description='Healthcheck')
    parser.add_argument(
        'addresses', metavar='address', type=str, default=[], nargs='*',
        help='one or more addresses to ping',
    )
    args = parser.parse_args()

    db = aiocheck.Database()
    if len(addresses) == 0:
        addresses = set(args.addresses)
    clear_console = lambda: os.system('cls' if os.name=='nt' else 'clear')

    def get_header(text):
        def wrap():
            result = ''
            for _ in range(len(text) + 4):
                result += '#'
            return result
        
        return f"{wrap()}\n# {text} #\n{wrap()}\n\nAddresses: {list(addresses)}\n"
    
    if len(addresses) == 0:
        select = ''
        while select != '0':
            inner_select = ''
            clear_console()
            print(f"{get_header('Menu')}\n[1] - Run\n[2] - Add Address\n[0] - Exit\n")
            select = input()
            
            if select == '0':
                sys.exit(0)
            
            if select == '1':
                if len(addresses) == 0:
                    input('No Address Provided ')
                else:
                    break
            
            if select == '2':
                clear_console()
                print(f"{get_header('Add Address')}")
                inner_select = input('Enter an Address: ')
                if len(inner_select) > 0:
                    addresses.add(inner_select)

    clear_console()
    sys.stdout.write(f"{get_header('Running')}\nPress CTRL+C to exit ")

    for address in addresses:
        db.insert(aiocheck.Host(address, db))

    loop = asyncio.get_event_loop()

    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass

if __name__ == "__main__":
    main()



"""
    aiocheck

    A python asyncio host health checker using native ping commands.

    Example:
    ```
    import aiocheck
    aiocheck.cli.main(['localhost'])
    ```

    :copyright: 2020 kruserr
    :license: MIT
"""

from aiocheck.Database import Database
from aiocheck.Host import Host
from aiocheck import cli



import setuptools


with open('README.md', 'r') as f:
    LONG_DESCRIPTION = f.read()

with open('VERSION', 'r') as f:
    VERSION = f.read()

setuptools.setup(
    name='aiocheck',
    version=VERSION,
    author='kruserr',
    description='A python asyncio host health checker using native ping commands',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url='https://github.com/kruserr/aiocheck',
    keywords='asyncio host health checker native ping',
    project_urls={
        'Documentation': 'https://github.com/kruserr/aiocheck/wiki',
        'Source': 'https://github.com/kruserr/aiocheck',
    },
    packages=setuptools.find_packages(
        where='src',
    ),
    package_dir={
        '': 'src',
    },
    install_requires=[],
    entry_points = {
        'console_scripts': ['aiocheck=aiocheck.cli:main'],
    },
    zip_safe=True,
    classifiers=[
        'Topic :: System :: Networking :: Monitoring',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)



import os
import json
import asyncio
import datetime


class Database():
    """
        Database class to handle access to data
        and access to file io.

        Example:
        ```
        db = Database()
        db.insert(Host('localhost', db))
        ```

        The Database supports async read and write
        from memory, and will persist to disk
        async every 5 seconds.
    """

    def __init__(self):
        self.__hosts = []
        self.__errors = []
        self.__last_errors = []
        self.__file_locked = False
        self.__errors_locked = False
        
        try:
            with open('.aiocheck_persist.json', 'r') as f:
                self.__errors = json.load(f)

            for item in self.__errors:
                item['timestamp'] = datetime.datetime.strptime(item['timestamp'], '%Y-%m-%d %H:%M:%S.%f')
        except Exception:
            pass

        asyncio.ensure_future(self.__persist_loop())
    
    def insert(self, host):
        """
            Insert host into database

            Example:
            ```
            db.insert(Host('localhost', db))
            ```
        """

        self.__hosts.append(host)

    def select(self, addr = None):
        """
            Read host from database

            Example:
            ```
            print(db.select())
            ```

            Supplied with addr, the host with that
            addr will be returned.

            Supplied with no addr, all hosts will
            be returned.

            if no hosts were found, None will be
            returned.
        """

        if addr is None:
            return self.__hosts
        
        for i in range(len(self.__hosts)):
            if self.__hosts[i].get_addr() == addr:
                return self.__hosts[i]
        return None
    
    def insert_error(self, host_as_json):
        """
            Insert error into database

            Example:
            ```
            db.insert_error(Host('localhost', db).get_json())
            ```

            Data is in json format, and is used to persist the
            database.
        """

        try:
            if self.__errors_locked:
                return
            
            self.__errors_locked = True

            if host_as_json in self.__errors:
                self.__errors_locked = False
                return

            match = False
            for i in range(len(self.__errors)):
                if self.__errors[i]['addr'] == host_as_json['addr']:
                    self.__errors[i] = host_as_json
                    match = True

            if not match:
                self.__errors.append(host_as_json)
            
            self.__errors = sorted(self.__errors, key = lambda i: i['timestamp'], reverse=True)
        except Exception as e:
            print(e)
        finally:
            self.__errors_locked = False

    def select_error(self, addr = None):
        """
            Read error from database

            Example:
            ```
            print(db.select_error())
            ```

            Supplied with addr, the error with that
            addr will be returned.

            Supplied with no addr, all error will
            be returned.

            if no error were found, {} will be
            returned.
        """

        if self.__errors_locked:
            return {}
        
        if addr is None:
            return self.__errors

        for error in self.__errors:
            if error['addr'] == addr:
                return error
        return {}

    def select_newest_errors(self):
        """
            Read newest errors from database

            Example:
            ```
            print(db.select_newest_errors())
            ```

            if no error were found, [] will be
            returned.
        """

        if self.__errors_locked:
            return []

        result = []
        check_addr = []
        for error in self.__errors:
            if error['addr'] not in check_addr:
                check_addr.append(error['addr'])
                result.append(error)
        return result

    def __error_equal(self, first, second):
        if first['addr'] != second['addr']:
            return False
        if first['alive'] != second['alive']:
            return False
        if first['timestamp'] != second['timestamp']:
            return False
        return True

    def __errors_equal(self):
        for new in self.__errors:
            for old in self.__last_errors:
                if new['addr'] == old['addr']:
                    if not self.__error_equal(new, old):
                        return False
        return True

    def __load_from_log(self):
        files = os.listdir(os.curdir)
        print(files)

    async def __persist_loop(self):
        while True:
            asyncio.ensure_future(self.__export())
            await asyncio.sleep(5)
    
    async def __export(self):
        try:
            if self.__file_locked:
                return

            if self.__errors_equal():
                self.__last_errors = self.__errors
                return

            self.__file_locked = True

            if len(self.__errors) > 0:
                with open('.aiocheck_persist.json', 'w') as f:
                    json.dump(self.__errors, f, indent=4, default=str)
        
            result = 'address, alive, timestamp\n'

            for error in self.__errors:
                result += f"{error['addr']}, {error['alive']}, {error['timestamp']}\n"

            if len(result) > 0:
                with open('aiocheck_log.csv', 'w') as f:
                    f.write(result)

            self.__last_errors = self.__errors
        except PermissionError:
            pass
        except Exception as e:
            print(e)
        finally:
            self.__file_locked = False
